﻿//-----------------------------------------------------------------------
// <copyright file="NodeCollection.cs" company="Telerik">
//     Copyright (c) Telerik.
// </copyright>
// <author>Miroslav Paskov</author>
//-----------------------------------------------------------------------

namespace HierarchyConverter
{
    using System;
    using System.Collections.ObjectModel;
    using System.ComponentModel;

    /// <summary>
    /// An observable collection that manages id relations as part of a ReverseHierarchyConverter.
    /// </summary>
    /// <typeparam name="TItem">The types of item stored in the collection.</typeparam>
    internal class NodeCollection<TItem> : ObservableCollection<TItem>
    {
        /// <summary>
        /// Holds a weak reference to the base collection.
        /// </summary>
        private WeakReference baseCollectionReference;

        /// <summary>
        /// Initializes a new instance of the NodeCollection class.
        /// </summary>
        /// <param name="baseCollection">The collection that manages this node collection, a ReverseHierarcyConverter.</param>
        /// <param name="nodeParentId">The id of the parent of all the items in the collection.</param>
        internal NodeCollection(IHierarchyBaseCollection<TItem> baseCollection, object nodeParentId)
        {
            this.baseCollectionReference = new WeakReference(baseCollection);
            this.ParentId = nodeParentId;
        }

        /// <summary>
        /// Gets the id of the node whose children this collection contains.
        /// </summary>
        internal object ParentId
        {
            get;
            private set;
        }

        /// <summary>
        /// Gets the base collection which has generated this node collection.
        /// </summary>
        internal IHierarchyBaseCollection<TItem> BaseCollection
        {
            get
            {
                return this.baseCollectionReference.Target as IHierarchyBaseCollection<TItem>;
            }
        }

        /// <summary>
        /// Removes all items from the collection.
        /// </summary>
        protected override void ClearItems()
        {
            foreach (var item in this)
            {
                this.StopTrackingItem(item);
            }

            var shouldClearBase = this.BaseCollection.OnNodeCollectionClearItems(this);
            if (shouldClearBase)
            {
                base.ClearItems();
            }
        }

        /// <summary>
        /// Inserts an item into the collection at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index at which <paramref name="item"/> should be inserted.</param>
        /// <param name="item">The object to insert.</param>
        protected override void InsertItem(int index, TItem item)
        {
            base.InsertItem(index, item);
            this.BaseCollection.OnNodeCollectionInsertItem(this, index, item);
            this.StartTrackingItem(item);
        }

        /// <summary>
        /// Removes the item at the specified index of the collection.
        /// </summary>
        /// <param name="index">The zero-based index of the element to remove.</param>
        protected override void RemoveItem(int index)
        {
            var oldItem = this[index];
            
            this.StopTrackingItem(oldItem);
            base.RemoveItem(index);
            this.BaseCollection.OnNodeCollectionRemoveItem(this, index, oldItem);
        }

        /// <summary>
        /// Replaces the element at the specified index.
        /// </summary>
        /// <param name="index">The zero-based index of the element to replace.</param>
        /// <param name="item">The new value for the element at the specified index.</param>
        protected override void SetItem(int index, TItem item)
        {
            var oldItem = this[index];

            this.StopTrackingItem(oldItem);
            base.SetItem(index, item);
            this.BaseCollection.OnNodeCollectionSetItem(this, index, oldItem, item);
            this.StartTrackingItem(item);
        }

        /// <summary>
        /// Called when a property on one of the items has changed.
        /// </summary>
        /// <param name="sender">The item where a property has changed.</param>
        /// <param name="e">The <see cref="System.ComponentModel.PropertyChangedEventArgs"/> instance containing the event data.</param>
        private void OnItemPropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            this.BaseCollection.OnNodeCollectionItemChanged(this, (TItem)sender, e);
        }

        /// <summary>
        /// Stops tracking property change in an item.
        /// </summary>
        /// <param name="item">The item that should not be tracked anymore.</param>
        private void StopTrackingItem(TItem item)
        {
            var observable = item as INotifyPropertyChanged;
            if (observable != null)
            {
                observable.PropertyChanged -= new PropertyChangedEventHandler(this.OnItemPropertyChanged);
            }
        }

        /// <summary>
        /// Starts lsitening for property changes in an item.
        /// </summary>
        /// <param name="item">The item which should be tracked.</param>
        private void StartTrackingItem(TItem item)
        {
            var observable = item as INotifyPropertyChanged;
            if (observable != null)
            {
                observable.PropertyChanged += new PropertyChangedEventHandler(this.OnItemPropertyChanged);
            }
        }
    }
}
