﻿//-----------------------------------------------------------------------
// <copyright file="ReverseHierarchyConverter.cs" company="Telerik">
//     Copyright (c) Telerik.
// </copyright>
// <author>Miroslav Paskov</author>
//-----------------------------------------------------------------------

namespace HierarchyConverter
{
	using System;
	using System.Collections;
	using System.Collections.Generic;
	using System.Collections.Specialized;
	using System.ComponentModel;
	using System.Diagnostics;
	using System.Globalization;
	using System.Linq;
	using System.Linq.Expressions;
	using System.Reflection;
	using System.Windows.Data;
	using X = System.Linq.Expressions.Expression;
	using System.Text;

	/// <summary>
	/// A converter that can bs used to reverse the hierarchy in a list of items so that it 
	/// transparently creates SL/WPF friendly a Parent -> Children hiearchy from a 
	/// Child -> Parent collection. 
	/// </summary>
	/// <typeparam name="TItem">The type of the items that make up the hiearchy.</typeparam>
	/// <typeparam name="TId">The type of the property that will be used as an id.</typeparam>
	public class ReverseHierarchyConverter<TItem, TId> : IValueConverter, INotifyPropertyChanged, IHierarchyBaseCollection<TItem>
	{
		/// <summary>
		/// References the original items source that the hierarchy was created from.
		/// </summary>
		private IEnumerable<TItem> originalItemsSource;

		/// <summary>
		/// A helper functions that returns the parent id of a given item.
		/// </summary>
		private Func<TItem, TId> getParentId;

		/// <summary>
		/// A helper function that sets the parent id of a given item.
		/// </summary>
		private Action<TItem, TId> setParentId;

		/// <summary>
		/// A helper function that returns the id of a given item.
		/// </summary>
		private Func<TItem, TId> getId;

		/// <summary>
		/// Contains a reference to an item whoose properties not be automatically changed. This is an optimizations for the cases where 
		/// the id properties are modifies by the hieararchy itself.
		/// </summary>
		private object suspendChangesOnItem;

		/// <summary>
		/// A flag specifying whether items should not be unassigned even if removed from the hierarchy.
		/// This is used to avoid unassigning an item when removing it due to parent property change.
		/// </summary>
		private bool suspendUnassign;

		/// <summary>
		/// Contains the children collections for a give id.
		/// </summary>
		private Dictionary<TId, NodeCollection<TItem>> childrenById = new Dictionary<TId, NodeCollection<TItem>>(128);

		/// <summary>
		/// Initializes a new instance of the ReverseHierarchyConverter class. When the default constructor is used the 
		/// converter needs to be subsequently initialized.
		/// </summary>
		public ReverseHierarchyConverter()
		{
		}

		/// <summary>
		/// Initializes a new instance of the ReverseHierarchyConverter class.
		/// </summary>
		/// <param name="idSelector">An expression that selects the property that will be used as an Id of the item.</param>
		/// <param name="parentIdSelector">An expression that will be used as a ParendId for the item.</param>
		/// <param name="rootItemParentIdValue">The value of the parent id that indicates the items is a root node.</param>
		/// <param name="unassignedItemParentId">The value of the parent id that indicates that the item currently has no parent.</param>
		public ReverseHierarchyConverter(Expression<Func<TItem, TId>> idSelector, Expression<Func<TItem, TId>> parentIdSelector, TId rootItemParentIdValue, TId unassignedItemParentId)
		{
			this.Initialize(idSelector, parentIdSelector, rootItemParentIdValue, unassignedItemParentId);
		}

		/// <summary>
		///     Raised when the value of one of the properties changes.
		/// </summary>
		public event PropertyChangedEventHandler PropertyChanged;

		/// <summary>
		/// Gets the ItemSource for the owner of the hieararchical items - a TreeView, Menu, PanelBar, etc.
		/// </summary>
		public IEnumerable<TItem> RootItems
		{
			get
			{
				return this.EnsureNodeCollection(this.RootParentIdValue);
			}
		}

		/// <summary>
		/// Gets the name of the property that contains the ParentId.
		/// </summary>
		public string ParentIdPropertyName
		{
			get;
			private set;
		}

		/// <summary>
		/// Gets the name of the property that will be used to retrieve the Id of the items.
		/// </summary>
		public string IdPropertyName
		{
			get;
			private set;
		}

		/// <summary>
		/// Gets the value that will be used for the parent id of all root.
		/// </summary>
		public TId RootParentIdValue
		{
			get;
			private set;
		}

		/// <summary>
		/// Gets the value for the items that are not part of the hierarchy.
		/// </summary>
		public TId UnassignedParentIdValue
		{
			get;
			private set;
		}

		/// <summary>
		/// Gets or sets the original flat list of items that will be used to generate the hierarchical data.
		/// </summary>
		public IEnumerable<TItem> ItemsSource
		{
			get
			{
				return this.originalItemsSource;
			}

			set
			{
				if (this.originalItemsSource != value)
				{
					this.CheckIsInitialized();
					var oldValue = this.originalItemsSource;
					this.originalItemsSource = value;
					this.OnItemsSourceChanged(oldValue, value);
					this.OnPropertyChanged("ItemsSource");
				}
			}
		}

		public IEnumerable<TItem> GetOrderedItemsSource()
		{
			foreach (var collection in childrenById.Values)
			{
				foreach (var item in collection)
				{
					yield return item;
				}
			}
		}

		public string GetPathForItem(TItem item)
		{
			var builder = new StringBuilder();
			builder.Append(getId(item));
			var parentId = getParentId(item);
			while (!object.Equals(parentId, this.UnassignedParentIdValue) && !object.Equals(parentId, this.RootParentIdValue))
			{
				item = originalItemsSource.FirstOrDefault(i => object.Equals(getId(i), parentId));
				builder.Insert(0, "|").Insert(0, getId(item).ToString());
				parentId = getParentId(item);
			}
			return builder.ToString();
		}

		/// <summary>
		/// Gets all the items that do not have a parent.
		/// </summary>
		public IEnumerable<TItem> UnassignedItems
		{
			get
			{
				return this.EnsureNodeCollection(this.UnassignedParentIdValue);
			}
		}

		/// <summary>
		/// Gets a value indicating whether the hiearchy converter has been initialized.
		/// </summary>
		public bool IsInitialized
		{
			get;
			private set;
		}

		/// <summary>
		/// Returns the children for a given item.
		/// </summary>
		/// <param name="parent">The item for which items should be retrieved.</param>
		/// <returns>A collection with the children for the given item.</returns>
		public IEnumerable<TItem> GetChildrenForItem(TItem parent)
		{
			var id = this.getId(parent);
			return this.EnsureNodeCollection(id);
		}

		/// <summary>
		/// The convert method that is expectedto be called when the binding of the ItemsSource property in the
		/// hioerarchical items control is applied.
		/// </summary>
		/// <param name="value">The value of the property that the converter will use.</param>
		/// <param name="targetType">The target type expected by the conversion, an IEnumerable is expected.</param>
		/// <param name="parameter">A parameter for the conversion, it will be ignored.</param>
		/// <param name="culture">The culture to use, it will be ignored.</param>
		/// <returns>A children colelction for the given item.</returns>
		public object Convert(object value, Type targetType, object parameter, CultureInfo culture)
		{
			if (value is TItem)
			{
				this.CheckIsInitialized();
				var result = this.GetChildrenForItem((TItem)value);
				var hierarchyAware = value as IHierarchyAware;
				if (hierarchyAware != null)
				{
					hierarchyAware.SetChildren(result, this);
				}

				return result;
			}

			return null;
		}

		/// <summary>
		/// The ConvertBack method is not supported by the HierarchyConverter.
		/// </summary>
		/// <param name="value">The value that is produced by the binding target.</param>
		/// <param name="targetType">The type to convert to.</param>
		/// <param name="parameter">The converter parameter to use.</param>
		/// <param name="culture">The culture to use in the converter.</param>
		/// <returns>
		/// A converted value. If the method returns null, the valid null value is used.
		/// </returns>
		public object ConvertBack(object value, Type targetType, object parameter, System.Globalization.CultureInfo culture)
		{
			throw new NotSupportedException("The ConvertBack method is not supported by the ReverseHierarchyConverter.");
		}

		/// <summary>
		/// Called by a child collection when an the INotifyPropertyChanged event of one of its items has been called.
		/// </summary>
		/// <param name="nodeCollection">The node collection whose item has changed.</param>
		/// <param name="item">The item that has changed.</param>
		/// <param name="args">The event args of the property change.</param>
		void IHierarchyBaseCollection<TItem>.OnNodeCollectionItemChanged(NodeCollection<TItem> nodeCollection, TItem item, PropertyChangedEventArgs args)
		{
			var changedPropertyName = args.PropertyName;

			// On parent id changed:
			if (changedPropertyName.Equals(this.ParentIdPropertyName))
			{
				this.suspendUnassign = true;
				try
				{
					nodeCollection.Remove(item);
				}
				finally
				{
					this.suspendUnassign = false;
				}

				this.EnsureNodeCollectionAndAdd(item, this.getParentId(item));
			}

			// On id changed:
			if (changedPropertyName.Equals(this.IdPropertyName))
			{
				throw new NotImplementedException(String.Format(CultureInfo.CurrentCulture, "Changing the {0}.{1} on an item is not allowed while in a hierarchy.", typeof(TItem).Name, this.IdPropertyName));
			}
		}

		/// <summary>
		/// Called when the children collection has been cleared.
		/// </summary>
		/// <param name="nodeCollection">The collection that has been cleared.</param>
		/// <returns>
		/// Returns whetehr the base.Clear() should be called.
		/// </returns>
		bool IHierarchyBaseCollection<TItem>.OnNodeCollectionClearItems(NodeCollection<TItem> nodeCollection)
		{
			var unasignedList = this.UnassignedItems as IList<TItem>;
			if (nodeCollection == unasignedList)
			{
				return false;
			}

			if (!this.suspendUnassign)
			{
				foreach (var item in nodeCollection.ToList())
				{
					this.UnassignItem(item);
					unasignedList.Add(item);
				}
			}

			return true;
		}

		/// <summary>
		/// Called when an item has been inserted in the children collection.
		/// </summary>
		/// <param name="nodeCollection">The children collection where an item has been inserted.</param>
		/// <param name="index">The index where an item has been inserted.</param>
		/// <param name="item">The item that has been inserted in the collection.</param>
		void IHierarchyBaseCollection<TItem>.OnNodeCollectionInsertItem(NodeCollection<TItem> nodeCollection, int index, TItem item)
		{
			var itemsSourceAsList = this.originalItemsSource as IList;
			if (itemsSourceAsList != null && !itemsSourceAsList.Contains(item))
			{
				itemsSourceAsList.Add(item);
			}

			this.AssignItem(item, (TId)nodeCollection.ParentId);
		}

		/// <summary>
		/// Called when an item has been removed from the collection.
		/// </summary>
		/// <param name="nodeCollection">The chidlren collection whoose item was removed.</param>
		/// <param name="index">The index of the removed item.</param>
		/// <param name="oldItem">The item that was removed.</param>
		void IHierarchyBaseCollection<TItem>.OnNodeCollectionRemoveItem(NodeCollection<TItem> nodeCollection, int index, TItem oldItem)
		{
			this.UnassignItem(oldItem);
		}

		/// <summary>
		/// Called when an item in the children collection has been replaced.
		/// </summary>
		/// <param name="nodeCollection">The collection where an item has been replaced.</param>
		/// <param name="index">The index of the replaced item.</param>
		/// <param name="oldItem">The item that was removed.</param>
		/// <param name="item">The item that was inserted.</param>
		void IHierarchyBaseCollection<TItem>.OnNodeCollectionSetItem(NodeCollection<TItem> nodeCollection, int index, TItem oldItem, TItem item)
		{
			this.UnassignItem(oldItem);

			// This is needed because change tracking is not happening.
			(this.UnassignedItems as IList).Add(oldItem);
			this.AssignItem(item, (TId)nodeCollection.ParentId);
		}

		/// <summary>
		/// Initializes this instance of the ReverseHierarchyConverter.
		/// </summary>
		/// <param name="idSelector">An expression that selects the property that will be used as an Id of the item.</param>
		/// <param name="parentIdSelector">An expression that will be used as a ParendId for the item.</param>
		/// <param name="rootItemParentIdValue">The value of the parent id that indicates the items is a root node.</param>
		/// <param name="unassignedItemParentIdValue">The value of the parent id that indicates that the item currently has no parent.</param>
		protected void Initialize(Expression<Func<TItem, TId>> idSelector, Expression<Func<TItem, TId>> parentIdSelector, TId rootItemParentIdValue, TId unassignedItemParentIdValue)
		{
			if (this.IsInitialized)
			{
				return;
			}

			if (object.Equals(rootItemParentIdValue, unassignedItemParentIdValue))
			{
				throw new ArgumentException("The rootItemParentIdValue and unassignedItemParenId values should not be equal");
			}

			if (idSelector == null)
			{
				throw new ArgumentNullException("idSelector");
			}

			if (parentIdSelector == null)
			{
				throw new ArgumentNullException("parentIdSelector");
			}

			this.UnassignedParentIdValue = unassignedItemParentIdValue;
			this.RootParentIdValue = rootItemParentIdValue;
			this.GenerateCompiledHelperMethods(idSelector, parentIdSelector);
			this.IsInitialized = true;
		}

		/// <summary>
		///     Called when the value of a property changes.
		/// </summary>
		/// <param name="propertyName">The name of the property that has changed.</param>
		protected void OnPropertyChanged(string propertyName)
		{
			if (this.PropertyChanged != null)
			{
				this.PropertyChanged(this, new PropertyChangedEventArgs(propertyName));
			}
		}

		/// <summary>
		/// Generats a method that can set the given property.
		/// </summary>
		/// <param name="property">The property for which a method will be generated.</param>
		/// <returns>A method that sets the given property.</returns>
		private static Action<TItem, TId> GenerateSetMethod(PropertyInfo property)
		{
			var itemParam = X.Parameter(typeof(TItem), "item");
			var idParam = X.Parameter(typeof(TId), "value");
			return X.Lambda<Action<TItem, TId>>(
						X.Call(itemParam, property.GetSetMethod(), idParam),
						itemParam,
						idParam).Compile();
		}

		/// <summary>
		/// Checks that the given expression is a simple accessor of a property and returns this property.
		/// </summary>
		/// <param name="selector">The expression that selects a property.</param>
		/// <returns>The PropertyInfo object reperesenting this property.</returns>
		private static PropertyInfo CheckSelectorAndRetrieveProperty(Expression<Func<TItem, TId>> selector)
		{
			var memberExpression = selector.Body as MemberExpression;
			if (memberExpression == null)
			{
				throw new ArgumentException("Selectors should be given in the form item => item.Property.");
			}

			var propertyInfo = memberExpression.Member as PropertyInfo;
			if (propertyInfo == null)
			{
				throw new ArgumentException("Only properties can be selected by the selectors.");
			}

			if (propertyInfo.GetSetMethod() == null)
			{
				throw new ArgumentException(String.Format(CultureInfo.CurrentCulture, "The {0} property does not have a public setter.", propertyInfo.Name));
			}

			return propertyInfo;
		}

		/// <summary>
		/// Returns the children collection for a given id and creates if not present.
		/// </summary>
		/// <param name="id">The id for which a collection will be geenrated.</param>
		/// <returns>Returns the collection for the given id.</returns>
		private NodeCollection<TItem> EnsureNodeCollection(TId id)
		{
			NodeCollection<TItem> result;
			if (!this.childrenById.TryGetValue(id, out result))
			{
				this.childrenById[id] = result = new NodeCollection<TItem>(this, id);
			}

			return result;
		}

		/// <summary>
		/// Checks whether the hierarchy has been initialized
		/// </summary>
		private void CheckIsInitialized()
		{
			if (!this.IsInitialized)
			{
				throw new InvalidOperationException("The HierarchyConverter must be initialized before it is used.");
			}
		}

		/// <summary>
		/// Called when the original ItemsSource for the hierarchy is changed. This items source should be used
		/// instead of the ItemsSource of the  respective control.
		/// </summary>
		/// <param name="oldValue">The old value for the ItemsSource.</param>
		/// <param name="newValue">The new value for the ItemsSource.</param>
		private void OnItemsSourceChanged(IEnumerable<TItem> oldValue, IEnumerable<TItem> newValue)
		{
			if (oldValue != null)
			{
				this.OnItemsSourceCleared(oldValue);
			}

			if (newValue != null)
			{
				this.OnItemsSourceAssigned(newValue);
			}

			this.OnPropertyChanged("RootItems");
			this.OnPropertyChanged("UnassignedItems");
		}

		/// <summary>
		/// Generats helper methods from the initializing expressions.
		/// </summary>
		/// <param name="idSelector">The selector for the id of an item.</param>
		/// <param name="parentIdSelector">The selector for the id of the parent of an item.</param>
		private void GenerateCompiledHelperMethods(Expression<Func<TItem, TId>> idSelector, Expression<Func<TItem, TId>> parentIdSelector)
		{
			var idProperty = CheckSelectorAndRetrieveProperty(idSelector);
			var parentIdProperty = CheckSelectorAndRetrieveProperty(parentIdSelector);

			this.getId = idSelector.Compile();
			this.getParentId = parentIdSelector.Compile();

			this.setParentId = GenerateSetMethod(parentIdProperty);

			this.ParentIdPropertyName = parentIdProperty.Name;
			this.IdPropertyName = idProperty.Name;
		}

		/// <summary>
		/// Creates all children collection. This is eager initialization and could be replaced by a lazy initialization later.
		/// </summary>
		private void CreateAllCollections()
		{
			foreach (var item in this.originalItemsSource)
			{
				var id = this.getParentId(item);
				this.EnsureNodeCollectionAndAdd(item, id);
			}
		}

		/// <summary>
		/// Makes sure a node colelction exists and adds an item to it.
		/// </summary>
		/// <param name="item">The ite to add to the collection.</param>
		/// <param name="id">The id for which a collection should be created or retrieved.</param>
		private void EnsureNodeCollectionAndAdd(TItem item, TId id)
		{
			NodeCollection<TItem> children;
			if (!this.childrenById.TryGetValue(id, out children))
			{
				this.childrenById[id] = children = new NodeCollection<TItem>(this, id);
			}
			if (!children.Contains(item))
			{
				children.Add(item);
			}
		}

		/// <summary>
		/// Clears the children collections without updating item parent ids.
		/// </summary>
		private void ClearAllCollections()
		{
			this.suspendUnassign = true;
			try
			{
				var rootCollection = this.EnsureNodeCollection(this.RootParentIdValue);
				rootCollection.Clear();
			}
			finally
			{
				this.suspendUnassign = false;
			}

			this.childrenById.Clear();
		}

		/// <summary>
		/// Called when the value of the items source has changed.
		/// </summary>
		/// <param name="newValue">The new items source.</param>
		private void OnItemsSourceAssigned(IEnumerable newValue)
		{
			var notifyCollectionChanged = newValue as INotifyCollectionChanged;
			if (notifyCollectionChanged != null)
			{
				notifyCollectionChanged.CollectionChanged += new NotifyCollectionChangedEventHandler(this.OnItemsSourceCollectionChanged);
			}

			this.CreateAllCollections();
		}

		/// <summary>
		/// Called when the original items source has been changed.
		/// </summary>
		/// <param name="oldValue">The old value of the items source for the hieararchy.</param>
		private void OnItemsSourceCleared(IEnumerable oldValue)
		{
			var notifyCollectionChanged = oldValue as INotifyCollectionChanged;
			if (notifyCollectionChanged != null)
			{
				notifyCollectionChanged.CollectionChanged -= new NotifyCollectionChangedEventHandler(this.OnItemsSourceCollectionChanged);
			}

			this.ClearAllCollections();
		}

		/// <summary>
		/// Called when the origina items source changes.
		/// </summary>
		/// <param name="sender">The itemssource that has changed.</param>
		/// <param name="e">Event arguments that represent the change.</param>
		private void OnItemsSourceCollectionChanged(object sender, NotifyCollectionChangedEventArgs e)
		{
			Debug.Assert(sender == this.originalItemsSource, "This method should only eb called for chnages in the original items collection.");

			if (e.Action == NotifyCollectionChangedAction.Add)
			{
				this.OnItemsChangedItemAdd((TItem)e.NewItems[0]);
			}
			else if (e.Action == NotifyCollectionChangedAction.Remove)
			{
				this.OnItemsChangedItemRemove((TItem)e.OldItems[0]);
			}
			else if (e.Action == NotifyCollectionChangedAction.Replace)
			{
				this.OnItemsChangedItemAdd((TItem)e.NewItems[0]);
				this.OnItemsChangedItemRemove((TItem)e.OldItems[0]);
			}
			else if (e.Action == NotifyCollectionChangedAction.Reset)
			{
				this.ClearAllCollections();
				this.CreateAllCollections();
			}
		}

		/// <summary>
		/// Called when an item has been removed fomr the original source.
		/// </summary>
		/// <param name="item">The item that has been removed from the original source.</param>
		private void OnItemsChangedItemRemove(TItem item)
		{
			var id = this.getId(item);
			var parentId = this.getParentId(item);
			NodeCollection<TItem> children;

			// Then remove the item:
			this.suspendChangesOnItem = item;
			if (this.childrenById.TryGetValue(parentId, out children))
			{
				children.Remove(item);
			}

			this.suspendChangesOnItem = null;
		}

		/// <summary>
		/// Called when an item has been added to the original items source.
		/// </summary>
		/// <param name="item">The item that has been added to the items source.</param>
		private void OnItemsChangedItemAdd(TItem item)
		{
			var parentId = this.getParentId(item);

			this.suspendChangesOnItem = item;
			this.EnsureNodeCollectionAndAdd(item, parentId);
			this.suspendChangesOnItem = null;
		}

		/// <summary>
		/// Sets the ParentId of the item to the given id.
		/// </summary>
		/// <param name="item">The item that wull be assigned to a new parent.</param>
		/// <param name="id">The id of the parent that this item will be assigned to.</param>
		private void AssignItem(TItem item, TId id)
		{
			if (!object.Equals(item, this.suspendChangesOnItem))
			{
				this.setParentId(item, id);
			}
		}

		/// <summary>
		/// Sets the ParendId of the item to indicate that it has been unassigned or is not part of the hierarchy.
		/// </summary>
		/// <param name="item">The item that will be unassigned.</param>
		private void UnassignItem(TItem item)
		{
			if (!this.suspendUnassign)
			{
				this.setParentId(item, this.UnassignedParentIdValue);
			}
		}
	}
}
